const { Logger } = require('twlv-logger');
const assert = require('assert');
const { readonly } = require('./lib/prop');
const { Identity } = require('twlv-identity');

const MODE_PLAIN = 0;
const MODE_ENCRYPTED = 1;

const ENCODING_BINARY = 0;
const ENCODING_STRING = 1;
const ENCODING_JSON = 2;

class Codec {
  constructor (identity) {
    assert(identity, 'Codec must define identity');

    readonly(this, 'identity', identity);
    readonly(this, 'logger', new Logger('twlv-swarm:codec'));
  }

  detectEncoding (payload) {
    if (payload === undefined || payload === null || payload === '') {
      return ENCODING_BINARY;
    }

    if (payload instanceof Buffer) {
      return ENCODING_BINARY;
    }

    if (typeof payload === 'string') {
      return ENCODING_STRING;
    }

    return ENCODING_JSON;
  }

  encode (payload, encoding) {
    assert(typeof encoding === 'number', 'Encoding must define as number');

    if (encoding === ENCODING_BINARY) {
      return payload;
    }

    if (encoding === ENCODING_JSON) {
      payload = JSON.stringify(payload);
    }

    return Buffer.from(payload, 'utf8');
  }

  decode (payload, encoding) {
    assert(typeof encoding === 'number', 'Encoding must define as number');
    assert(payload instanceof Buffer, 'Cannot decode non-buffer payload');

    if (encoding === ENCODING_BINARY) {
      return payload;
    }

    if (encoding === ENCODING_JSON) {
      return JSON.parse(payload);
    }

    return payload.toString('utf8');
  }

  pack ({ kind, module, command, mode = MODE_ENCRYPTED, payload }, identity) {
    assert(typeof kind === 'string' && kind.length === 1, 'Kind of module must be one byte string');
    assert(typeof module === 'string', 'Module must be string');
    assert(typeof command === 'string', 'Command must be string');
    assert(typeof mode === 'number', 'Mode must be number');

    // remove payload when payload is empty buffer
    if ((payload instanceof Buffer || typeof payload === 'string') && !payload.length) {
      payload = undefined;
    }

    // force to mode plain when payload is empty or identity not specified
    if (!payload || !identity) {
      mode = MODE_PLAIN;
    }

    // detect encoding and encode payload for identity
    let encoding = this.detectEncoding(payload);
    payload = this.encode(payload, encoding);

    // when encrypted mode then sign and encrypt
    if (mode === MODE_ENCRYPTED) {
      payload = Identity.signThenEncrypt(payload, this.identity, identity);
    }

    let buffer = Buffer.alloc(34 + (payload ? payload.length : 0));
    buffer.write(kind, 0, 1);
    buffer.write(module, 1, 23);
    buffer.write(command, 24, 8);
    buffer.writeUInt8(mode, 32);
    buffer.writeUInt8(encoding, 33);
    if (payload) {
      payload.copy(buffer, 34);
    }

    return buffer;
  }

  unpack (buffer, identity, keys = [ 'kind', 'module', 'command', 'mode', 'payload' ]) {
    let kind;
    let module;
    let command;
    let mode;
    let payload;

    if (keys.indexOf('kind') !== -1) {
      kind = buffer.toString('utf8', 0, 1);
    }

    if (keys.indexOf('module') !== -1) {
      module = buffer.toString('utf8', 1, 23).replace(/\u0000/g, '');
    }

    if (keys.indexOf('command') !== -1) {
      command = buffer.toString('utf8', 24, 32).replace(/\u0000/g, '');
    }

    if (keys.indexOf('mode') !== -1) {
      mode = buffer.readUInt8(32);
    }

    if (keys.indexOf('payload') !== -1) {
      let encoding = buffer.readUInt8(33);

      payload = buffer.slice(34);

      if (payload.length) {
        if (mode === MODE_ENCRYPTED) {
          if (!identity) {
            throw new Error('Cannot unpack encrypted message without peer identity');
          }

          payload = Identity.decryptThenVerify(payload, this.identity, identity);
        }

        payload = this.decode(payload, encoding);
      }
    }

    return { kind, module, command, mode, payload };
  }
}

module.exports = {
  Codec,
  MODE_PLAIN,
  MODE_ENCRYPTED,
  ENCODING_BINARY,
  ENCODING_STRING,
  ENCODING_JSON,
};
