const { Peer } = require('../../peer');
const { readonly, hidden } = require('../../lib/prop');
const { Logger } = require('twlv-logger');
const { DEFAULT_TIMEOUT, KIND, MODULE, CMD_ADV } = require('./constants');
const { HandshakeSocketUnit } = require('./socket-unit');
const { HandshakeRelayUnit } = require('./relay-unit');

class Handshake {
  constructor (timeout = DEFAULT_TIMEOUT) {
    readonly(this, 'timeout', timeout);
    hidden(this, 'units', []);

    hidden(this, 'swarm');
    hidden(this, 'logger');
    hidden(this, '__onConnect');
  }

  up (swarm) {
    this.swarm = swarm;
    this.logger = new Logger('twlv-swarm:controller:handshake', swarm.address.substr(0, 4));

    this.__onConnect = session => {
      let unit = this.getRelayUnit(session.address);
      if (unit) {
        let resolve = unit._resolveCallback;
        this.removeRelayUnit(unit);
        resolve(session);
      }
    };
    swarm.on('connect', this.__onConnect);
  }

  down () {
    if (this.swarm) {
      this.swarm.removeListener('connect', this.__onConnect);
      this.__onConnect = undefined;
      this.swarm = undefined;
      this.logger = undefined;
    }
  }

  async handshakeWithSocket (socket, initiate = false) {
    this.logger.log('Trying handshake with socket ...');
    let { timeout, swarm: { advertisement, codec } } = this;
    let unit = new HandshakeSocketUnit({ advertisement, codec, timeout, socket, initiate });
    let session = await unit.createSession();
    this.logger.log('Session handshake done');
    return session;
  }

  async handshakeWithRelay (address, initiate = false) {
    this.logger.log('Trying handshake with relay ...');
    let unit = this.addRelayUnit(address, initiate);
    let session = await unit.createSession();
    this.removeRelayUnit(unit);
    this.logger.log('Session handshake done');
    return session;
  }

  addRelayUnit (address, initiate) {
    let { timeout, swarm } = this;
    let unit = new HandshakeRelayUnit({ swarm, timeout, address, initiate });
    this.units.push(unit);

    return unit;
  }

  removeRelayUnit (unit) {
    let index = this.units.indexOf(unit);
    if (index !== -1) {
      this.units.splice(index, 1);
    }
  }

  getRelayUnit (address) {
    return this.units.find(unit => unit.address === address);
  }

  async _onMessage ({ address, kind, module, command, payload }) {
    if (kind !== KIND || module !== MODULE) {
      return;
    }

    let unit = this.getRelayUnit(address);
    if (unit) {
      unit._onMessage({ address, kind, module, command, payload });
      return;
    }

    if (command === CMD_ADV) {
      try {
        let peer = new Peer(payload);
        await this.swarm.connect(peer, true);
      } catch (err) {
        this.logger.warn('Failed when trying backconnect, fallback to relay');

        let unit = this.addRelayUnit(address, false);
        let session = await unit.createSession(payload);
        this.removeRelayUnit(unit);

        this.swarm._addSession(session);
      }

      return;
    }

    throw new Error(`Unimplemented handshake command ${command}`);
  }
}

module.exports = { Handshake };
