const { Peer } = require('../../peer');
const { Session } = require('../../session');
const { KIND, MODULE, CMD_ADV, CMD_ACK } = require('./constants');
const { MODE_PLAIN } = require('../../codec');
const { Logger } = require('twlv-logger');

class HandshakeSocketUnit {
  constructor ({ advertisement, codec, timeout, socket, initiate }) {
    this.session = new Session({ codec, socket, initiate });
    this.advertisement = advertisement;
    this.timeout = timeout;
    this.logger = new Logger('twlv-swarm:controller:handshake:socket-unit');
  }

  async createSession () {
    await this.session.write({
      mode: MODE_PLAIN,
      kind: KIND,
      module: MODULE,
      command: CMD_ADV,
      payload: this.advertisement,
    });

    if (!this.session.handshaken) {
      await new Promise((resolve, reject) => {
        let timeout = setTimeout(() => reject(new Error('Handshake timeout')), this.timeout);

        this.session.on('readable', () => {
          let message = this.session.read();
          while (message) {
            try {
              let done = this._onMessage(message);
              if (done) {
                clearTimeout(timeout);
                return resolve();
              }

              message = this.session.read();
            } catch (err) {
              this.session.end();
              clearTimeout(timeout);
              return reject(err);
            }
          }
        });

        this.session.on('end', () => {
          this.session.removeAllListeners('readable');
          this.session.removeAllListeners('end');
          clearTimeout(timeout);

          reject(new Error('Handshake rejected'));
        });
      });
    }

    return this.session;
  }

  _onMessage ({ kind, module, command, payload }) {
    if (kind !== KIND || module !== MODULE) {
      return;
    }

    if (command === CMD_ADV) {
      if (payload.address === this.advertisement.address) {
        throw new Error('Connect to self');
      }

      this.peer = new Peer(payload);
      this.session.write({
        mode: MODE_PLAIN,
        kind: KIND,
        module: MODULE,
        command: CMD_ACK,
      });

      this._acking = true;
    } else if (command === CMD_ACK && this.peer) {
      this.session.removeAllListeners('readable');
      this.session.removeAllListeners('end');

      this._acked = true;
    }

    if (this._acking && this._acked) {
      this.session.handshakeWith(this.peer);
      return true;
    }
  }
}

module.exports = { HandshakeSocketUnit };
