const assert = require('assert');
const { MODE_PLAIN } = require('../codec');
const { Logger } = require('twlv-logger');
const { hidden } = require('../lib/prop');

const KIND = '!';
const MODULE = 'relay';
const CMD_SEND = 's';

class Relay {
  constructor (swarm) {
    hidden(this, 'swarm');
    hidden(this, 'logger');
  }

  up (swarm) {
    this.swarm = swarm;
    this.logger = new Logger('twlv-swarm:controller:relay', swarm.address.substr(0, 4));
    this._clock = new RelayClock();
    this._clocks = {};
  }

  down () {
    if (this.swarm) {
      this.swarm = undefined;
      this.logger = undefined;
    }
  }

  relay ({ address, ttl, kind, module, command, payload }, identity) {
    payload = this._createRelayPayload({ address, ttl, kind, module, command, payload }, identity);

    this.swarm.sessions.forEach(session => {
      if (!session.handshaken) {
        return;
      }
      session.write({ mode: MODE_PLAIN, kind: KIND, module: MODULE, command: CMD_SEND, payload });
    });
  }

  _createRelayPayload ({ address, ttl = 30, kind, module, command, payload }, identity) {
    assert(ttl <= 30, 'TTL must be lower than 30');

    let payloadBuffer = this.swarm.codec.pack({ kind, module, command, payload }, identity);
    let buffer = Buffer.alloc(22 + payloadBuffer.length);
    let clock = this._clock.next();

    buffer.writeUInt8(ttl, 0);
    buffer.writeUInt8(clock, 1);
    Buffer.from(this.swarm.address, 'hex').copy(buffer, 2);
    Buffer.from(address, 'hex').copy(buffer, 12);
    payloadBuffer.copy(buffer, 22);

    return buffer;
  }

  readTTL (payload) {
    return payload.readUInt8(0);
  }

  readClock (payload) {
    return payload.readUInt8(1);
  }

  readFrom (payload) {
    return payload.toString('hex', 2, 12);
  }

  readTo (payload) {
    return payload.toString('hex', 12, 22);
  }

  decrementTTL (payload) {
    let ttl = this.readTTL(payload);
    payload.writeUInt8(ttl - 1);
  }

  readMessage (payload) {
    let message = this.swarm.codec.unpack(payload.slice(22));
    message.address = this.readFrom(payload);
    return message;
  }

  _onMessage ({ address, kind, module, command, mode, payload }) {
    let ttl = this.readTTL(payload);
    if (ttl <= 0) {
      return;
    }

    let from = this.readFrom(payload);
    if (from === this.swarm.address) {
      return;
    }

    let srcClock = this._getClock(from);
    let clock = this.readClock(payload);
    if (srcClock.happenBefore(clock)) {
      srcClock.update(clock);
    } else {
      return;
    }

    let to = this.readTo(payload);
    if (to === this.swarm.address) {
      let message = this.readMessage(payload);
      // this.logger.log('Inbound relay message %o', message);
      return this.swarm._onMessage(message);
    }

    this.decrementTTL(payload);

    this._propagate(address, payload);
  }

  _propagate (address, payload) {
    this.swarm.sessions.forEach(session => {
      if (session.address === address || !session.handshaken) {
        return;
      }

      session.write({ kind: KIND, module: MODULE, command: CMD_SEND, payload });
    });
  }

  _getClock (address) {
    let clock = this._clocks[address] = this._clocks[address] || new RelayClock();
    return clock;
  }
}

class RelayClock {
  constructor (value = -1) {
    this.value = value;
    this.min = 0;
    this.median = 256 / 2;
    this.max = 256 - 1;
  }

  next () {
    this.value = this.value === this.max ? 0 : this.value + 1;
    return this.value;
  }

  happenBefore (remoteValue) {
    assert(typeof remoteValue === 'number', 'Remote value must be number');
    assert(remoteValue <= this.max, `Invalid remoteValue, ${remoteValue}`);

    if (this.value < this.min) {
      return true;
    }

    if (this.value < remoteValue) {
      return true;
    }

    if (this.value > this.median && this.value < remoteValue + this.median) {
      return true;
    }

    return false;
  }

  update (remoteValue) {
    this.value = remoteValue;
  }
}

module.exports = { Relay };
