const uuidv1 = require('uuid/v1');
const { readonly } = require('../lib/prop');
const { Logger } = require('twlv-logger');
const assert = require('assert');
const WebSocket = require('ws');

class Client {
  constructor ({ socket, module }) {
    this.id = uuidv1();

    assert(socket, 'Client needs parameter socket');

    readonly(this, 'logger', new Logger('twlv-swarm:daemon:client'));
    readonly(this, 'module', module);
    readonly(this, 'socket', socket);

    if (module) {
      module.on('message', this._onModuleMessage.bind(this));
    }
    socket.on('message', this._onWireMessage.bind(this));
  }

  _onModuleMessage (message) {
    this.write('message', message);
  }

  async _onWireMessage (frame) {
    let { command, payload } = JSON.parse(frame);

    this.logger.log('IPC<server>: %o', command);

    if (this.module) {
      switch (command) {
        case 'send': {
          let id = payload.id;
          try {
            await this.module.send({
              address: payload.address,
              command: payload.command,
              payload: payload.payload,
            });
            this.write('!send', { id });
          } catch (err) {
            this.logger.warn('Caught while send:', err.message);
            this.write('!send', { id, error: err.message });
          }
          break;
        }
        case 'request': {
          let id = payload.id;
          try {
            await this.module.send({
              address: payload.address,
              command: '!req',
              payload: {
                id,
                command: payload.command,
                payload: payload.payload,
              },
            });
          } catch (err) {
            this.logger.log('Client "request" caught "%s"', err.message);
            this.write('message', {
              address: payload.address,
              command: '!res',
              payload: { id, error: err.message },
            });
          }
          break;
        }
        case 'relay': {
          this.module.relay(payload);
          break;
        }
        case 'broadcast': {
          this.module.broadcast(payload);
          break;
        }
        case 'lookup': {
          let { id, address } = payload;
          try {
            let value = await this.module.lookup(address);
            this.write('!lookup', { id, value });
          } catch (err) {
            this.write('!lookup', { id, error: err.message });
          }
          break;
        }
        default:
          this.logger.warn(`Unimplemented client command: ${command}`);
      }
    }
  }

  write (command, payload) {
    if (this.socket.readyState === WebSocket.OPEN) {
      this.socket.send(JSON.stringify({ command, payload }));
      return;
    }

    if (this.socket.readyState === WebSocket.CONNECTING || this.socket.readyState === WebSocket.CLOSING) {
      this.logger.warn('Failed write to socket with readyState', this.socket.readyState);
    }
  }
}

module.exports = { Client };
