const { Config } = require('twlv-config');
const { Logger } = require('twlv-logger');
const fs = require('fs-extra');
const path = require('path');
const { Swarm } = require('../');
const { readonly, hidden } = require('../lib/prop');
const http = require('http');
const WebSocket = require('ws');
const { Client } = require('./client');
const { KIND_DICTIONARY } = require('../module');

class Daemon {
  constructor (config = new Config(), swarm = new Swarm()) {
    this.logger = new Logger('twlv-swarm:daemon');

    this.config = config;
    this.swarm = swarm;
    this.status = 0;

    readonly(this, 'useLocking', Boolean(this.config.get('useLocking')));
    readonly(this, 'dataDir', path.resolve(this.config.get('dataDir') || path.join(process.cwd(), 'data')));
    readonly(this, 'lockFile', path.join(this.dataDir, 'twlvd.lock'));
    readonly(this, 'libDir', path.resolve(this.config.get('libDir') || path.resolve(Swarm.HOME, '..')));

    hidden(this, '_ipcServer');
    hidden(this, 'clients', []);
  }

  get port () {
    return this._ipcServer ? this._ipcServer.address().port : 0;
  }

  get ipcUrl () {
    return `ws://localhost:${this.port}`;
  }

  get pid () {
    return process.pid;
  }

  get address () {
    return this.swarm.address;
  }

  async start (server) {
    this.logger.log('Starting daemon ...');

    await this._assertLock();

    this.logger.info(`Daemon PID: ${this.pid}`);
    this.logger.info(`Daemon config file: ${this.config.file}`);
    this.logger.info(`Daemon datadir: ${this.dataDir}`);
    this.logger.info(`Daemon libdir: ${this.libDir}`);
    this.logger.info(`Daemon locking: ${this.useLocking}`);

    await this._bootIPC(server);

    await this._bootSwarm();
    await this._bootModules();

    await this._writeLock();

    this.status = 1;
  }

  async stop () {
    if (!this.status) {
      return;
    }

    this.status = 0;

    this.logger.log('Stopping daemon ...');

    await this._debootModules();
    await this._debootSwarm();
    await this._debootIPC();

    await this._removeLock();

    if (this.config.file) {
      await this.config.persist();
      this.logger.info('Configuration persisted');
    }

    this.logger.info('Daemon stopped');

    return true;
  }

  async _bootSwarm () {
    this._swarmAdvertisementListener = advertisement => this.broadcast('advertisement', advertisement);
    this.swarm.on('advertisement', this._swarmAdvertisementListener);

    await this.swarm.start();

    this.logger.info(`Swarm address: ${this.swarm.address}`);
  }

  async _debootSwarm () {
    this.swarm.removeAllListeners('advertisement', this._swarmAdvertisementListener);
    this._swarmAdvertisementListener = null;

    await this.swarm.stop();
  }

  async _bootModules () {
    let configs = this.config.get('modules');
    if (configs) {
      await Promise.all(configs.map(async config => {
        if (config.autostart) {
          let { kind, name, options } = config;

          if (this.swarm.getModule(kind, name)) {
            throw new Error('Module already exists');
          }
          let module = this._createModule({ kind, name, options });
          await this.swarm.addModule(module);
        }
      }));
    }
  }

  _debootModules () {
    let configs = this.config.get('modules');
    if (configs) {
      let newConfigs = [];
      configs.forEach(config => {
        let { kind, name, options } = config;
        let module = this.swarm.getModule(kind, name);
        let autostart = Boolean(module);
        newConfigs.push({ kind, name, options, autostart });
      });

      this.config.set('modules', newConfigs);
    }
  }

  async _bootIPC (server) {
    this.useExternalServer = Boolean(server);

    if (!this.useExternalServer) {
      let port = this.config.get('port');
      server = http.Server((req, res) => res.end(JSON.stringify({ name: 'twlv' })));
      await new Promise(resolve => server.listen(port, resolve));
    }

    this._ipcServer = server;

    const sock = new WebSocket.Server({ server });
    sock.on('connection', this._onIpcIncoming.bind(this));
    this.sock = sock;

    this.logger.info(`IPC listening at ws://localhost:${this.port}`);
  }

  _debootIPC () {
    let server = this._ipcServer;
    this._ipcServer = null;

    if (this.useExternalServer) {
      return;
    }

    server.close();

    this.logger.info('IPC stopped successfully');
  }

  _createModule ({ kind, name, options }) {
    if (this.swarm.getModule(kind, name)) {
      throw new Error('Module already exists');
    }

    let ModuleClass = this._findModuleClass(kind, name);
    return new ModuleClass(options);
  }

  _findModuleClass (kind, name) {
    // try builtin
    try {
      let spec = path.join(Swarm.HOME, KIND_DICTIONARY[kind], name);
      return require(spec);
    } catch (err) {
    }

    try {
      let spec = path.join(this.libDir, `twlv-${KIND_DICTIONARY[kind]}-${name}`);
      return require(spec);
    } catch (err) {
      throw new Error(`Can't find module ${KIND_DICTIONARY[kind]}:${name}
  You might want to install:

  npm install -g twlv-${KIND_DICTIONARY[kind]}-${name}
`);
    }
  }

  _createExternalModule ({ kind, name, version }) {
    if (this.swarm.getModule(kind, name)) {
      throw new Error('Module already exists');
    }

    let ModuleClass = require(`./external/${KIND_DICTIONARY[kind]}`);
    return new ModuleClass({ name, version });
  }

  _onIpcIncoming (socket) {
    socket.on('message', async frame => {
      try {
        let { command, payload } = JSON.parse(frame);
        if (command !== 'connect') {
          throw new Error('Invalid connection');
        }

        let { address, networkId, advertisement } = this.swarm;
        let { privateKey } = this.swarm.identity;

        let module;
        if (payload) {
          module = this._createExternalModule(payload);
          await this.swarm.addModule(module);
        }

        socket.send(JSON.stringify({
          command: '!connect',
          payload: { address, privateKey, networkId, advertisement },
        }));
        socket.removeAllListeners('message');

        let client = new Client({ socket, module });
        socket.on('close', async () => {
          socket.removeAllListeners();

          let { module } = client;
          await this.swarm.removeModule(module);

          let index = this.clients.indexOf(client);
          if (index !== -1) {
            this.clients.splice(index, 1);
          }
        });
        this.clients.push(client);
      } catch (err) {
        this.logger.error('IPC Incoming caught "%s"', err.message);
        socket.close();
      }
    });
  }

  _addClient (client) {
  }

  async _removeClient (client) {

  }

  broadcast (command, payload) {
    this.clients.forEach(client => client.write(command, payload));
  }

  async _assertLock () {
    if (!this.useLocking) {
      return;
    }

    await fs.ensureDir(this.dataDir);

    if (await fs.exists(this.lockFile)) {
      throw new Error(`Cannot start already run daemon at data dir ${this.dataDir}`);
    }
  }

  async _writeLock () {
    if (!this.useLocking) {
      return;
    }

    let { pid, port } = this;
    await fs.writeFile(this.lockFile, JSON.stringify({ pid, port }));
  }

  async _removeLock () {
    if (!this.useLocking) {
      return;
    }

    await fs.remove(this.lockFile);

    this.logger.log('Lock removed');
  }
}

module.exports = { Daemon };
