const assert = require('assert');
const { Logger } = require('twlv-logger');
const { vreadonly, hidden } = require('./lib/prop');
const { Request } = require('./request');
const { EventEmitter } = require('events');

const KIND_APP = 'a';
const KIND_CHANNEL = 'c';
const KIND_DISCOVERY = 'd';
const KIND_CONTROL = '!';

const KIND_DICTIONARY = {
  a: 'app',
  c: 'channel',
  d: 'discovery',
  '!': 'control',
};

class Module extends EventEmitter {
  constructor (kind, name, version) {
    super();

    assert(kind, 'Module must defined kind');
    assert(name, 'Module must defined name');
    assert(name.length <= 23, 'Module name maximum is 23 chars');

    vreadonly(this, 'kind', kind);
    vreadonly(this, 'name', name);
    vreadonly(this, 'version', version || '0.0.0');

    assert('up' in this, `Module ${this} must implement up`);
    assert('down' in this, `Module ${this} must implement down`);
    assert('onMessage' in this, `Module ${this} must implement onMessage`);

    hidden(this, '_facade');
    hidden(this, 'requests');
    hidden(this, 'logger', new Logger(`twlv-swarm:${KIND_DICTIONARY[this.kind]}-${this.name}`));
  }

  get facade () {
    assert(this._facade, `Module ${this} does not have facade, maybe not up yet`);

    return this._facade;
  }

  get address () {
    return this.facade.address;
  }

  get networkId () {
    return this.facade.networkId;
  }

  get advertisement () {
    return this.facade.advertisement;
  }

  get status () {
    return this._facade ? 1 : 0;
  }

  async _start (facade) {
    assert(facade, 'Start must include facade access');

    if (this.status > 0) {
      return;
    }

    this._facade = facade;
    this.logger = new Logger(`twlv-swarm:${KIND_DICTIONARY[this.kind]}-${this.name}`, facade.address.substr(0, 4));
    this.requests = [];

    await this.up();
  }

  async _stop () {
    if (this.status <= 0) {
      return;
    }

    let req;
    while ((req = this.requests.shift())) {
      req.resolve();
    };

    await this.down();

    this.logger = null;
    this._facade = null;
  }

  lookup (address) {
    return this.facade.lookup(address);
  }

  broadcast ({ command, payload }) {
    return this.facade.broadcast({ command, payload });
  }

  send ({ address, command, payload }) {
    return this.facade.send({ address, command, payload });
  }

  request ({ address, command, payload, timeout }) {
    assert(this.onRequest, `Module ${this} must implement onRequest in order to use request`);

    let req = new Request(() => {
      let index = this.requests.indexOf(req);
      if (index !== -1) {
        this.requests.splice(index, 1);
      }
    }, timeout);

    this.requests.push(req);
    return req.run(async req => {
      await this.send({
        address,
        command: '!req',
        payload: {
          id: req.id,
          command,
          payload,
        },
      });
    });
  }

  relay ({ address, command, payload }, identity) {
    return this.facade.relay({ address, command, payload });
  }

  dump () {
    let { kind, name, version, status } = this;
    return { kind, name, version, status };
  }

  toString () {
    return `${KIND_DICTIONARY[this.kind]}:${this.name}`;
  }

  async _onMessage ({ address, command, payload }) {
    if (command === '!req') {
      let { id, command, payload: req } = payload;
      try {
        let value = await this.onRequest({ address, id, command, payload: req });
        this.send({ address, command: '!res', payload: { id, value } });
      } catch (err) {
        this.logger.error('Request caught "%s"', err.message);
        let error = `Remote: ${err.message}`;
        this.send({ address, command: '!res', payload: { id, error } });
      }
    } else if (command === '!res') {
      let req = this.requests.find(req => req.id === payload.id);
      if (req) {
        let { value, error } = payload;
        if (error) {
          req.reject(new Error(error));
        } else {
          req.resolve(value);
        }
      }
    } else {
      this.onMessage({ address, command, payload });
    }
  }
}

module.exports = { Module, KIND_APP, KIND_CHANNEL, KIND_CONTROL, KIND_DISCOVERY, KIND_DICTIONARY };
