const assert = require('assert');
const lpstream = require('length-prefixed-stream');
const { Logger } = require('twlv-logger');
const { hidden, readonly } = require('./lib/prop');
const { Readable } = require('stream');
const { KIND_APP, KIND_CHANNEL, KIND_DISCOVERY, KIND_CONTROL } = require('./module');

class Session extends Readable {
  constructor ({ codec, socket, initiate = false }) {
    super({ objectMode: true });

    assert(codec, 'Session must define codec');
    assert(socket, 'Session must define socket as stream');

    readonly(this, 'codec', codec);
    readonly(this, 'socket', socket);
    readonly(this, 'initiate', initiate);
    readonly(this, 'logger', new Logger('twlv-swarm:session'));

    hidden(this, 'peer');
    hidden(this, 'encoder', lpstream.encode());
    hidden(this, 'decoder', lpstream.decode());

    this.encoder.pipe(socket).pipe(this.decoder);
    socket.on('end', this._onSocketEnd.bind(this));

    this.decoder.on('readable', () => {
      let frame = this.decoder.read();
      if (frame) {
        let message = this._frameToMessage(frame);
        this.push(message);
      }
    });

    hidden(this, '_readableState', this._readableState);
  }

  get address () {
    return this.peer ? this.peer.address : undefined;
  }

  get identity () {
    return this.peer ? this.peer.identity : undefined;
  }

  get connected () {
    return Boolean(this.socket);
  }

  get handshaken () {
    return this.connected && Boolean(this.peer);
  }

  advertise (kind, module) {
    if (!this.peer) {
      return false;
    }

    if (kind === KIND_CONTROL) {
      return true;
    }

    switch (kind) {
      case KIND_APP:
        return this.peer.apps.indexOf(module) !== -1;
      case KIND_CHANNEL:
        return this.peer.channels.indexOf(module) !== -1;
      case KIND_DISCOVERY:
        return this.peer.discoveries.indexOf(module) !== -1;
    }
  }

  dump () {
    let { address, connected, handshaken } = this;
    return {
      address,
      connected,
      handshaken,
    };
  }

  _frameToMessage (frame) {
    try {
      let { kind, module, command, mode, payload } = this.codec.unpack(frame, this.identity);
      let address = this.address;
      return { address, kind, module, command, mode, payload };
    } catch (err) {
      this.logger.error('Session message unpack caught "%s"', err.message);
    }
  }

  _onSocketEnd () {
    this.peer = undefined;
    this.destroy();
  }

  _read () {
    // noop
  }

  handshakeWith (peer) {
    this.peer = peer;
  }

  write ({ mode, kind, module, command, payload }) {
    let message = this.codec.pack({ mode, kind, module, command, payload }, this.identity);
    this.encoder.write(message);
  }

  end () {
    if (!this.socket) {
      return;
    }

    this.socket.end();
    this.encoder.end();
    this.decoder.end();
  }
}

module.exports = { Session };
