const { Peer } = require('./peer');
const { Identity } = require('twlv-identity');
const { EventEmitter } = require('events');
const assert = require('assert');
const Memory = require('twlv-storage-memory');
const assertion = require('./lib/assertion');
const facade = require('./lib/facade');
const { Logger } = require('twlv-logger');
const { Codec } = require('./codec');
const { KIND_APP, KIND_CHANNEL, KIND_DISCOVERY, KIND_CONTROL } = require('./module');
const { hidden, readonly } = require('./lib/prop');
const { Handshake, Relay } = require('./controller');
const packageJson = require('./package.json');
const sleep = require('es7-sleep');

const LOOKUP_TIMEOUT = 10000;

/**
 * Object to initiate swarm network or join to network
 */
class Swarm extends EventEmitter {
  static get VERSION () {
    return packageJson.version;
  }

  static get HOME () {
    return __dirname;
  }

  /**
   *
   * @param {object} options Options to initiate swarm
   *
   * Available options:
   *
   * networkId - The network id
   * storage - Storage instance
   */
  constructor ({ networkId = '0', storage = new Memory(), timeout } = {}) {
    super();

    assert(networkId, 'Swarm must define networkId');
    assert(storage, 'Swarm must define storage');

    this.peers = [];
    this.sessions = [];
    this.modules = [];

    readonly(this, 'networkId', networkId);
    hidden(this, 'bootPeers', []);
    hidden(this, 'storage', storage);
    hidden(this, 'logger', new Logger('twlv-swarm'));

    readonly(this, 'handshakeController', new Handshake(timeout));
    readonly(this, 'relayController', new Relay());

    hidden(this, '_lookupUnits', []);

    hidden(this, '_debounceStateChanged');
    this.on('module-start', this._updateAdvertisement.bind(this));
    this.on('module-stop', this._updateAdvertisement.bind(this));
  }

  _updateAdvertisement () {
    let urls = [];
    let apps = [];
    let channels = [];
    let discoveries = [];

    this.modules.forEach(module => {
      switch (module.kind) {
        case KIND_APP:
          apps.push(module.name);
          break;
        case KIND_CHANNEL:
          channels.push(module.name);
          // TODO: revisit generate urls to include current connected socket urls
          let result = module.getUrls();
          if (result) {
            urls = urls.concat(result);
          }
          break;
        case KIND_DISCOVERY:
          discoveries.push(module.name);
          break;
      }
    });

    let { address, publicKey } = this;

    this.advertisement = { address, publicKey, urls, apps, channels, discoveries };

    this.logger.log('Advertisement updated %o', { address, channels, discoveries, apps });

    this.broadcast({ kind: KIND_CONTROL, module: 'update', command: 'adv', payload: this.advertisement });

    this.emit('advertisement', this.advertisement);
  }

  get address () {
    return this.status ? this.identity.address : undefined;
  }

  get publicKey () {
    return this.status ? this.identity.publicKey : undefined;
  }

  get status () {
    return this.identity ? 1 : 0;
  }

  addBootPeer (peer) {
    assert(typeof peer === 'string' || peer instanceof Peer, 'Boot peer must be string or instanceof peer');

    this.bootPeers.push(peer);
  }

  getChannelByUrl (url) {
    let proto = url.split(':').shift();
    return this.modules.find(module => module.kind === KIND_CHANNEL && module.proto === proto);
  }

  async send ({ address, kind, module, command, payload }) {
    assert(typeof address === 'string', 'Address must be string');
    assertion.kind(kind);

    assert(kind === KIND_CONTROL || this.getModule(kind, module), 'Cannot send to unregistered module');

    let peer;
    let session = await this.get(address);
    if (!session) {
      peer = await this.lookup(address);
      assert(peer, `Peer ${address} not found`);

      try {
        session = await this.connect(peer);
      } catch (err) {
        this.logger.log('Failed to establish connection', err.message);
      }
    }

    assert(session, `Session for ${address} not found`);

    if (!session.advertise(kind, module)) {
      throw new Error('Cannot send message of unadvertised module');
    }

    return session.write({ kind, module, command, payload });
  }

  broadcast ({ kind, module, command, payload }) {
    assertion.kind(kind);

    assert(kind === KIND_CONTROL || this.getModule(kind, module), 'Cannot send to unregistered module');

    this.sessions.forEach(session => {
      if (!session.handshaken) {
        return;
      }

      if (!session.advertise(kind, module)) {
        return;
      }

      session.write({ kind, module, command, payload });
    });
  }

  relay ({ address, ttl, kind, module, command, payload }) {
    assertion.kind(kind);
    let session = this.get(address);
    this.relayController.relay({ address, ttl, kind, module, command, payload }, session ? session.identity : undefined);
  }

  get (address) {
    assert(typeof address === 'string', 'Address must be string');

    return this.sessions.find(session => session.address === address);
  }

  /**
   * Lookup peer by its address
   *
   * The function will try to find from connected peers first.
   *
   * When peer with specified address not found, then lookup to discovery methods.
   *
   * @param {string} address
   */
  async lookup (address) {
    this.logger.log(`Looking up address=${address}...`);

    // lookup from available local peers
    let peer = this.peers.find(peer => peer.address === address);
    if (peer) {
      return peer;
    }

    // FIXME: reimplement using lookup unit to avoid double lookup for the same
    // address or timeout leaking
    // when not found, try to lookup remotely by discovery
    let discoveryCount = this.modules.filter(module => module.kind === KIND_DISCOVERY).length;
    let finishedCount = 0;
    if (discoveryCount) {
      let definition = await new Promise(resolve => {
        let timeout = setTimeout(() => {
          this.logger.warn('Lookup timeout');
          resolve();
        }, LOOKUP_TIMEOUT);

        this._lookupUnits.push({ timeout, resolve });
        let resolveAndClean = (definition) => {
          let index = this._lookupUnits.indexOf(timeout);
          if (index !== -1) {
            this._lookupUnits.splice(index, 1);
          }
          clearTimeout(timeout);
          resolve(definition);
        };

        this.modules.forEach(async module => {
          if (module.kind !== KIND_DISCOVERY) return;

          try {
            this.logger.log(`Trying lookup using ${module}...`);
            let definition = await module.onLookup(address);
            if (definition) {
              this.logger.log(`Discovery ${module} found ${definition}`);
              resolveAndClean(definition);
            } else {
              finishedCount++;
              if (finishedCount >= discoveryCount) {
                resolveAndClean();
              }
            }
          } catch (err) {
            this.logger.error('Discovery lookup caught "%s"', err.message);
          }
        });
      });

      if (definition) {
        peer = this._putPeer(definition);
      }
    }

    // if (!peer) {
    //   this.logger.warn(`Failed to lookup ${address}`);
    // }
    return peer;
  }

  _putPeer (newPeer) {
    assert(newPeer instanceof Peer, 'Peer to put must be instance of peer');

    let peer = this.peers.find(peer => peer.address === newPeer.address);

    if (peer) {
      peer.update(newPeer);
    } else {
      this.peers.push(newPeer);
      peer = newPeer;
    }

    this.logger.log('Peer size', this.peers.length);
    return peer;
  }

  async connect (definition, forceDirect = false) {
    assert(this.status, 'Cant connect on stopped swarm');
    assert(definition, 'Connect must define definition');

    this.logger.log('Connecting to %s...', definition);

    if (typeof definition === 'string') {
      let session = await this._connectByUrl(definition);
      this.logger.log('Session connected to %s', session.peer);
      this.logger.log('Session size', this.sessions.length);
      return session;
    } else if (definition instanceof Peer) {
      let { urls } = definition;
      for (let url of urls) {
        let channel = this.getChannelByUrl(url);
        if (!channel) continue;

        try {
          let session = await this._connectByUrl(url);
          this.logger.log('Session connected to %s', session.peer);
          this.logger.log('Session size', this.sessions.length);
          return session;
        } catch (err) {
          this.logger.log(`Connect to ${url} caught "%s"`, err.message);
        }
      }

      if (!forceDirect) {
        this.logger.log('Failed to establish direct connection, trying connect use relay');
        try {
          let session = await this._connectByRelay(definition);
          this.logger.log('Session connected use relay to', session.peer.address);
          this.logger.log('Session size', this.sessions.length);
          return session;
        } catch (err) {
          this.logger.log('Failed to establish relay connection, cause "%s"', err.message);
        }
      }

      throw new Error(`Failed to connect to ${definition}`);
    } else {
      throw new Error(`Failed to connect to unknown definition`);
    }
  }

  async _connectByUrl (url) {
    let channel = this.getChannelByUrl(url);
    assert(channel, `Dial failed, suitable channel not found for ${url}`);
    let socket = await channel.connect(url);
    let session = await this._addSessionWithSocket(socket, true);
    return session;
  }

  async _connectByRelay (peer, initiate) {
    let session = await this._addSessionWithRelay(peer.address, true);
    return session;
  }

  async disconnect (session) {
    session.removeAllListeners();
    await session.end();
    this.emit('disconnect', session);

    let index = this.sessions.indexOf(session);
    if (index !== -1) {
      this.sessions.splice(index, 1);
    }

    this.logger.log('Session size', this.sessions.length);
  }

  async start () {
    if (this.status) {
      return;
    }

    await this._bootIdentity();

    // this._updateAdvertisement();

    await this._bootModules();
    await this._bootControllers();
    await this._bootSessions();

    this.logger.log('Swarm started');
    this.emit('start');
  }

  async stop () {
    if (!this.status) {
      return;
    }
    await this._debootSessions();
    await this._debootControllers();
    await this._debootModules();
    await this._debootIdentity();

    this.logger.log('Swarm stopped');
    this.emit('stop');
  }

  async _bootControllers () {
    await this.handshakeController.up(this);
    await this.relayController.up(this);
  }

  async _debootControllers () {
    await this.handshakeController.down();
    await this.relayController.down();
  }

  async _bootSessions () {
    await Promise.all(this.bootPeers.map(async url => {
      try { await this.connect(url); } catch (err) {}
    }));
  }

  async _debootSessions () {
    await Promise.all(this.sessions.map(session => this.disconnect(session)));
    this.sessions = [];

    // TODO: workaround lookup units
    await this._lookupUnits.forEach(({ timeout, resolve }) => {
      clearTimeout(timeout);
      resolve();
    });
    this._lookupUnits = [];
  }

  async _onMessage ({ address, kind, module, command, mode, payload }) {
    // TODO: implement rate limit

    // this.logger.log(`Inbound %o`, { address, kind, module, command, mode, payload });

    this.emit('message', { address, kind, module, command, payload });

    try {
      if (kind === KIND_CONTROL) {
        if (module === 'relay') {
          await this.relayController._onMessage({ address, kind, module, command, mode, payload });
          return;
        } else if (module === 'handshake') {
          await this.handshakeController._onMessage({ address, kind, module, command, mode, payload });
          return;
        } else if (module === 'update') {
          if (command === 'adv') {
            let newPeer = new Peer(payload);
            this.logger.log('Update advertisement of %s', newPeer);
            this._putPeer(newPeer);
            return;
          }
        } else if (module === 'mock') {
          // for testing purpose only
          return;
        }

        throw new Error(`Unimplemented handler for control message ${module}:${command}`);
      }

      let theModule = this.getModule(kind, module);
      if (theModule) {
        await theModule._onMessage({ address, command, payload });
      }
    } catch (err) {
      this.logger.error('Message routing "%s"', err.message);
    }
  }

  getModule (kind, name) {
    return this.modules.find(module => module.kind === kind && module.name === name);
  }

  getApp (name) {
    this.logger.warn(`DEPRECATED #getApp(), use #getModule()`);
    return this.getModule(KIND_APP, name);
  }

  getChannel (name) {
    this.logger.warn(`DEPRECATED #getChannel(), use #getModule()`);
    return this.getModule(KIND_CHANNEL, name);
  }

  getDiscovery (name) {
    this.logger.warn(`DEPRECATED #getDiscovery(), use #getModule()`);
    return this.getModule(KIND_DISCOVERY, name);
  }

  async addModule (module) {
    if (this.getModule(module.kind, module.name)) {
      throw new Error(`Module ${module} already exists`);
    }

    this.modules.push(module);

    if (this.status) {
      await module._start(facade(this, { module }));
      this.emit('module-start', module);
      this.logger.log(`Module ${module} started`);
    }
  }

  async removeModule (module) {
    let index = this.modules.indexOf(module);
    if (index !== -1) {
      this.modules.splice(index, 1);
      await module._stop();
      this.emit('module-stop', module);
      this.logger.log(`Module ${module} stopped`);
    }
  }

  async _bootModules () {
    await Promise.all(this.modules.map(async module => {
      try {
        await module._start(facade(this, { module }));
        this.emit('module-start', module);
        this.logger.log(`Module ${module} started`);
      } catch (err) {
        this.logger.warn(`Failed to start ${module}`);
        this.logger.log(`Failed to start ${module}`, err);
      }
    }));
  }

  async _debootModules () {
    await Promise.all(this.modules.map(async module => {
      try {
        await module._stop();
        this.emit('module-stop', module);
        this.logger.log(`Module ${module} stopped`);
      } catch (err) {
        this.logger.warn(`Failed to stop ${module}`);
        this.logger.log(`Failed to stop ${module}`, err);
      }
    }));
  }

  async incoming (socket) {
    if (!socket.pipe) {
      throw new Error('Incoming must be define socket');
    }

    try {
      await this._addSessionWithSocket(socket);
    } catch (err) {
      // TODO: notify socket about error, must define error messaging
      this.logger.error('Incoming "%s"', err.message);
    }
  }

  async _addSessionWithSocket (socket, initiate = false) {
    let session = await this.handshakeController.handshakeWithSocket(socket, initiate);
    this._addSession(session);
    await sleep();
    return session;
  }

  async _addSessionWithRelay (address, initiate = false) {
    let session = await this.handshakeController.handshakeWithRelay(address, initiate);
    this._addSession(session);
    await sleep();
    return session;
  }

  _addSession (session) {
    session.on('data', this._onMessage.bind(this));
    session.on('end', () => this.disconnect(session));

    session.peer = this._putPeer(session.peer);
    this.emit('connect', session);
    this.sessions.push(session);
  }

  async initIdentity () {
    this.identity = new Identity();
    await this.storage.write('privkey.pem', this.identity.privateKey);
  }

  async _bootIdentity () {
    if (await this.storage.exists('privkey.pem')) {
      this.identity = new Identity(await this.storage.read('privkey.pem'));
    } else {
      await this.initIdentity();
    }

    this.logger = new Logger('twlv-swarm', this.identity.address.substr(0, 4));
    this.codec = new Codec(this.identity);
  }

  _debootIdentity () {
    this.identity = null;
    this.codec = null;
  }
}

module.exports = { Swarm };
